﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.ComponentModel;
using System.Collections.ObjectModel;
using System.Windows.Data;
using System.Windows.Input;
using Buzz.MachineInterface;
using BuzzGUI.Interfaces;
using BuzzGUI.Common;
using System.Xml;

namespace videosync
{
    [MachineDecl(Name = "pac videosync 2", ShortName = "videosync", Author = "Pierre Cusa", MaxTracks = 8)]
    public class SyncMachine : IBuzzMachine
    {
        IBuzzMachineHost host;
        public SyncGUI Gui = null;
        public IEnumerable<IMenuItem> Commands { get; private set; }

        [ParameterDecl]
        public bool Dummy { get; set; }

        MenuItemVM guiMenuItem;
        MenuItemVM helpMenuItem;

        public Collection<TempoChange> TempoChanges = new Collection<TempoChange>();
        public string VideoPath { get; set; }
        public bool ManageTempos = false;
        public int VideoVolume = 0; // 0-100 (libvlc standard)
        public string VLCPath;
        public int VLCVout;
        float tpbFactor = 60000 / Global.Buzz.TPB;

        bool lastPlaying = false;
        bool prePlay = false;
        int lastPlayPos = 0;
        int buzzBpm = 126;

        public struct SyncInfo
        {
            public int Tick;
            public int Ms;
            public int Bpm;
        }

        static SyncInfo currentSync = new SyncInfo { Tick = -1, Ms = 0, Bpm = Global.Buzz.BPM };
        public SyncInfo CurrentSync
        {
            // calculate current video position and bpm from current buzz tick
            get
            {
                int buzzTickTime = Global.Buzz.Song.PlayPosition;
                if (currentSync.Tick == buzzTickTime)
                    return currentSync;

                int videoMsTime = 0;
                int lastBpm = buzzBpm; // todo find out why it segv's if i use global.buzz.bpm
                if (ManageTempos)
                {
                    int lastTick = 0;
                    lastBpm = TempoChanges[0].Bpm;
                    foreach (var tempoChange in TempoChanges)
                    {
                        if (tempoChange.Tick > buzzTickTime + 1)
                            break;
                        videoMsTime += Tick2ms(tempoChange.Tick - lastTick, lastBpm);
                        lastTick = tempoChange.Tick;
                        lastBpm = tempoChange.Bpm;
                    }
                    videoMsTime += Tick2ms(buzzTickTime - lastTick, buzzBpm);
                }
                else
                {
                    videoMsTime = Tick2ms(buzzTickTime, buzzBpm);
                }
                currentSync = new SyncInfo { Tick = buzzTickTime, Ms = videoMsTime, Bpm = lastBpm };
                return currentSync;
            }
        }

        public class DelegateCommand : ICommand
        {
            private readonly Action _execute;
            public event EventHandler CanExecuteChanged;

            public DelegateCommand(Action execute)
            { _execute = execute; }
            public bool CanExecute(object parameter)
            { return true; }
            public void Execute(object parameter)
            { _execute(); }
        }

        // constructor
        public SyncMachine(IBuzzMachineHost host)
        {
            this.host = host;

            guiMenuItem = new MenuItemVM();
            guiMenuItem.ID = 0;
            guiMenuItem.Text = "Video Settings";
            guiMenuItem.Command = new DelegateCommand(() => host.Machine.DoubleClick());

            helpMenuItem = new MenuItemVM();
            helpMenuItem.ID = 1;
            helpMenuItem.Text = "Help";
            helpMenuItem.Command = new DelegateCommand(() => host.Machine.ShowHelp());

            VLCVout = 0;
            Commands = new List<IMenuItem>(new IMenuItem[] { guiMenuItem, helpMenuItem });
            try
            {
                var doc = new XmlDocument();
                doc.Load(Global.BuzzPath + "\\Gear\\Generators\\pac videosync config.xml");
                var vlcPathNode = doc.SelectSingleNode("//vlcpath/text()");
                if (vlcPathNode != null)
                    VLCPath = vlcPathNode.Value;
                var vlcOutNode = doc.SelectSingleNode("//vlcout/text()");
                if (vlcOutNode != null)
                    VLCVout = int.Parse(vlcOutNode.Value);
            }
            catch (System.IO.FileNotFoundException)
            {
                // yeh whtvr man
            }
            TempoChanges.Add(new TempoChange { Tick = 0, Bpm = 126 });
            buzzBpm = Global.Buzz.BPM;
            Global.Buzz.Song.PropertyChanged += songPropertyChanged;
            Global.Buzz.PropertyChanged += buzzPropertyChanged;
        }

        public void Work()
        {
            if (Gui != null && Gui.vidIsOpen)
            {
                if (host.SubTickInfo.CurrentSubTick == 0 && host.SubTickInfo.PosInSubTick == 0)
                {
                    // a tick is starting
                    // set player volume on every tick because libvlc is being a prick
                    Gui.SetVolume(VideoVolume);
                    if (prePlay)
                    {
                        // real play is starting
                        prePlay = false;
                        Gui.Play();
                    }
                }

                if (Global.Buzz.Playing && !lastPlaying)
                    // preplay is starting (real play after the current tick finished)
                    prePlay = true;

                if (!Global.Buzz.Playing && lastPlaying)
                    // stopping
                    Gui.Pause();

                if (Global.Buzz.Song.PlayPosition != lastPlayPos && Global.Buzz.Song.PlayPosition != lastPlayPos + 1)
                {
                    // skipping is now
                    Gui.Skip(CurrentSync.Ms);
                }

                // if (prePlay && host.SubTickInfo.CurrentSubTick == host.SubTickInfo.SubTicksPerTick - 1)
                    // change the tempo in preplay's last subtick ? no

                lastPlaying = Global.Buzz.Playing;
                lastPlayPos = Global.Buzz.Song.PlayPosition;
            }
        }

        void buzzPropertyChanged(object sender, PropertyChangedEventArgs e)
        {
            if (e.PropertyName == "BPM")
            {
                buzzBpm = Global.Buzz.BPM;
                if (!Global.Buzz.Playing && !ManageTempos && Gui != null && Gui.vidIsOpen)
                    Gui.Skip(CurrentSync.Ms);
            }
            if (e.PropertyName == "TPB")
                tpbFactor = 60000 / Global.Buzz.TPB;
        }

        // this is still wrong because i can't change the bpm for the current tick.
        void songPropertyChanged(object sender, PropertyChangedEventArgs e)
        {
            if (!ManageTempos || e.PropertyName != "PlayPosition" || TempoChanges.Count == 0)
                return;
            if (Global.Buzz.BPM != CurrentSync.Bpm && CurrentSync.Bpm > 16 && CurrentSync.Bpm < 500)
                Global.Buzz.BPM = CurrentSync.Bpm;
        }

        public class StateData
        {
            public string StateVideoPath;
            public bool StateManageTempos = false;
            public int StateVideoVolume = 0;
            public Collection<TempoChange> stateTempoChanges;
        }

        public StateData MachineState {
            // save song etc
            get
            {
                return new StateData
                {
                    StateVideoPath = VideoPath,
                    stateTempoChanges = TempoChanges,
                    StateManageTempos = ManageTempos,
                    StateVideoVolume = VideoVolume
                };
            }
            // load song etc
            set
            {
                VideoPath = value.StateVideoPath;
                TempoChanges = value.stateTempoChanges;
                ManageTempos = value.StateManageTempos;
                VideoVolume = value.StateVideoVolume;
            }
        }

        public int Tick2ms(int time, int bpm)
        {
            return (int)(((float)time / (float)bpm) * tpbFactor);
        }

        public class TempoChange
        {
            int tick, bpm;
            public int Tick { get { return tick; } set { tick = value; currentSync.Tick = -1; } }
            public int Bpm { get { return bpm; } set { bpm = value; currentSync.Tick = -1; } }
            public TempoChange()
            {
                Tick = 0;
                Bpm = 126;
            }
        }
    }

    [MachineGUIFactoryDecl(PreferWindowedGUI = true, IsGUIResizable = true)]
    public class MachineGUIFactory : IMachineGUIFactory
    {
        public IMachineGUI CreateGUI(IMachineGUIHost host) { return new SyncGUI(); }
    }
}